"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
const experimental_utils_1 = require("@typescript-eslint/experimental-utils");
const tsutils_1 = require("tsutils");
const typescript_1 = __importStar(require("typescript"));
const common_1 = require("./common");
/**
 * An ESLint rule to ban unsafe type assertions.
 */
const noUnsafeTypeAssertion = {
    meta: {
        type: "problem",
        docs: {
            category: "Possible Errors",
            description: "Bans unsafe type assertions.",
            recommended: "error",
            url: "https://github.com/danielnixon/eslint-plugin-total-functions",
        },
        messages: {
            errorStringGeneric: "This type assertion is not type-safe.",
            errorStringPreferAs: "Use the 'as' syntax for type assertions.",
        },
        schema: [],
    },
    create: (context) => {
        const parserServices = experimental_utils_1.ESLintUtils.getParserServices(context);
        const checker = parserServices.program.getTypeChecker();
        const isUnsafe = (rawDestinationType, rawSourceType, seenTypes = []) => {
            // eslint-disable-next-line functional/no-conditional-statement
            if (rawSourceType.flags & typescript_1.TypeFlags.Any ||
                rawSourceType.flags & typescript_1.TypeFlags.Unknown) {
                // Asserting any or unknown to anything else is always unsafe.
                return true;
            }
            const { destinationType, sourceType } = common_1.filterTypes(rawDestinationType, rawSourceType);
            // eslint-disable-next-line functional/no-conditional-statement
            if (destinationType === undefined || sourceType === undefined) {
                return false;
            }
            return destinationType.getProperties().some((destinationProperty) => {
                const destinationPropertyType = common_1.symbolToType(destinationProperty, checker);
                const sourceProperty = sourceType.getProperty(destinationProperty.name);
                const sourcePropertyType = sourceProperty !== undefined
                    ? common_1.symbolToType(sourceProperty, checker)
                    : undefined;
                const destinationPropertyIsOptional = destinationProperty.flags & typescript_1.default.SymbolFlags.Optional;
                const destinationTypeParts = destinationPropertyType !== undefined
                    ? tsutils_1.unionTypeParts(destinationPropertyType)
                    : [];
                const destinationPropertyAllowsUndefined = destinationTypeParts.some((t) => t.flags & typescript_1.default.TypeFlags.Undefined);
                const sourcePropertyIsUndefined = sourcePropertyType === undefined ||
                    tsutils_1.unionTypeParts(sourcePropertyType).some((t) => t.flags & typescript_1.default.TypeFlags.Undefined);
                const nextSeenTypes = seenTypes.concat([
                    { destinationType, sourceType },
                ]);
                const isUnsafePropertyAssignment = sourcePropertyIsUndefined &&
                    !destinationPropertyAllowsUndefined &&
                    !destinationPropertyIsOptional;
                // This is an unsafe assignment if...
                return (
                // we're not in an infinitely recursive type,
                seenTypes.every((t) => t.destinationType !== destinationType &&
                    t.sourceType !== sourceType) &&
                    // and this is an unsafe property assignment, or
                    (isUnsafePropertyAssignment ||
                        // this is unsafe recursively
                        (destinationPropertyType !== undefined &&
                            sourcePropertyType !== undefined &&
                            isUnsafe(destinationPropertyType, sourcePropertyType, nextSeenTypes))));
            });
        };
        const reportUnsafe = (node
        // eslint-disable-next-line functional/no-return-void
        ) => {
            // eslint-disable-next-line functional/no-conditional-statement
            if (node.typeAnnotation.type === experimental_utils_1.AST_NODE_TYPES.TSTypeReference &&
                node.typeAnnotation.typeName.type === experimental_utils_1.AST_NODE_TYPES.Identifier &&
                node.typeAnnotation.typeName.name === "const") {
                // Always allow `as const`.
                return;
            }
            // The right hand side of the "as".
            const destinationNode = parserServices.esTreeNodeToTSNodeMap.get(node);
            const rawDestinationType = checker.getTypeAtLocation(destinationNode);
            // The left hand side of the "as".
            const sourceNode = destinationNode.expression;
            const rawSourceType = checker.getTypeAtLocation(sourceNode);
            // eslint-disable-next-line functional/no-conditional-statement
            if (isUnsafe(rawDestinationType, rawSourceType)) {
                // eslint-disable-next-line functional/no-expression-statement
                context.report({
                    node: node,
                    messageId: "errorStringGeneric",
                });
            }
        };
        return {
            TSTypeAssertion: reportUnsafe,
            TSAsExpression: reportUnsafe,
        };
    },
};
exports.default = noUnsafeTypeAssertion;
